from pathlib import Path
import os
import glob
import shutil
import logging

class SubPaths:
    """
    Represents the various directories and paths associated with a single sub-experiment

    - log
    - result
    - original (for re-ranking
    - conf
    """

    _LIBREC_PROPERTIES_FILE = 'librec.properties'
    DEFAULT_LOG_FILENAME = "librec.log"

    _path_dict = None

    _prop_dict = {'log': 'dfs.log.dir',
#                 'split': 'dfs.split.dir',
                 'result': 'dfs.result.dir',
                  'conf': 'dfs.config.dir'}

    _sub_dirs = ['conf', 'log', 'result', 'original']

    subexp_name = None

    def __init__(self, base, subexp_name, create=True):
        self._path_dict = {}
        self.subexp_name = subexp_name

        subexp_path = base / subexp_name
        self.set_path('subexp', subexp_path)

        status_path = subexp_path / '.status'
        self.set_path('status', status_path)

        for subdir in self._sub_dirs:
            subdir_path = subexp_path / subdir
            self.set_path(subdir, subdir_path)

        if create:
            logging.info("Creating subexperiment: {}", subexp_name)
            subexp_path.mkdir(exist_ok=True)
            for subdir in self._sub_dirs:
                self.get_path(subdir).mkdir(exist_ok=True)

    def get_path(self, type):
        if type in self._path_dict:
            return self._path_dict[type]
        else:
            return None

    def get_librec_properties_path(self):
        return self.get_path('conf') / self._LIBREC_PROPERTIES_FILE

    def get_path_str(self, type):
        return self.get_path(type).as_posix()

    def get_path_platform(self, type):
        return str(self.get_path(type))

    def get_path_prop(self, type):
        return self._prop_dict[type]

    def set_path(self, type, path):
        self._path_dict[type] = path

    def set_path_from_string(self, type, path_str):
        self._path_dict[type] = Path(path_str)

    # Assumes path is set up
    def add_to_config(self, config, type):
        prop_name = self.get_path_prop(type)
        prop_val = self.get_path_str(type)
        config[prop_name] = prop_val

    def results2original(self):
        original_path = self.get_path('original')
        result_path = self.get_path('result')
        shutil.rmtree(original_path)
        original_path.mkdir()
        files = glob.glob((result_path / '*').as_posix())
        for file in files:
            shutil.copy2(file, original_path)

    def original2results(self):
        original_path = self.get_path('original')
        result_path = self.get_path('result')
        files = glob.glob((original_path / '*').as_posix())
        for file in files:
            shutil.copy2(file, result_path)
        shutil.rmtree(original_path)




from pathlib import Path
import hashlib
import inspect
import librec_auto
from librec_auto.librec_auto.util import SubPaths
from librec_auto.librec_auto.util.utils import force_path
from collections import OrderedDict

class Files:
    """
    This class encapsulates the file system for librec-auto.

    The files for librec-auto can be stored in a number of possible locations:
    - global location: determined by the install location of librec-auto
    - rules-specific location: determined by the location of a rules directory [Not implemented]
    - user-specific location: associated with the user's home directory [Not implemented]
    - experiment-specific location: in a directory associated with a specific experiment. See sub_paths.py for a class
      that handles the sub-experiment-specific files.

    These are the types of files that are managed:
    - configuration files (XML). Stores information about how experiments are configured.
    - properties files (key=value format). Generated from configuration files for input to librec
    - split files (UIR or UIRT format). Generated by separating input rules files into different test/training splits
    - experiment logs (csv files). Content is algorithm-specific. Generated by a run of an experiment

    NOTE: User and rules directories are currently unused.
    """

    _global_dir_path = None
    _sub_path_dict = None
    _config_dir_path = None
    _rules_dir_path = None
    #_res_dir_path = None
    _split_dir_path = None
    _jar_dir_path = None
    _post_dir_path = None

    _prop_file_name = None

    _DEFAULT_GLOBAL_DIR_STR = inspect.getfile(librec_auto) # This needs to be tested.
#    _DEFAULT_GLOBAL_DIR_STR = "."

    _DEFAULT_CONFIG_DIR_NAME = "conf"
    _DEFAULT_RULES_DIR_NAME = "rules"
    _DEFAULT_RES_DIR_NAME = "result"
    _DEFAULT_SPLIT_DIR_NAME = "data/split"
    _DEFAULT_JAR_DIR_NAME = "librec_auto/jar"
    _DEFAULT_POST_DIR_NAME = "post"
    _EXP_DIR_PATTERN = "exp{:05d}"

    _DEFAULT_PROP_FILE_NAME = "librec.properties"
    _DEFAULT_LA_JAR = "auto.jar"
    _DEFAULT_LR_JAR = "librec.jar"
    _DEFAULT_RULES_FILE = "librec_auto/rules/element-rules.xml"
    LOG_PATH = "../log/librec.log"
    DEFAULT_CONFIG_FILENAME = "config.xml"

    # 2019-11-22 RB Do we use this anywhere?
    # _DEFAULT_CACHE_FILENAME = ".cache"

    def __init__(self):
        self._config_dir_path = Path(self._DEFAULT_CONFIG_DIR_NAME)
        self._rules_dir_path = Path(self._DEFAULT_RULES_DIR_NAME)
        #self._res_dir_path = Path(self._DEFAULT_RES_DIR_NAME)
        self._split_dir_path = Path(self._DEFAULT_SPLIT_DIR_NAME)
        self._jar_dir_path = Path(self._DEFAULT_JAR_DIR_NAME)
        self._post_dir_path = Path(self._DEFAULT_POST_DIR_NAME)
        self._sub_path_dict = OrderedDict()

        module_init_path = Path(Files._DEFAULT_GLOBAL_DIR_STR).parent
        self._global_path = module_init_path.parent
        # maybe_user_path = Path(Files._DEFAULT_USER_PATH_STR)
        # if maybe_user_path.is_dir():
        #     self._user_path = maybe_user_path

    def get_global_path(self): return self._global_path

    def get_jar_path(self): return self._jar_dir_path

    def get_exp_path(self): return self._exp_path

    def get_split_path(self): return self._exp_path / self._split_dir_path

    def get_config_path(self): return self._exp_path / self._config_dir_path / self._config_file_name

    def get_post_path(self): return self._exp_path / self._post_dir_path

    def set_global_path(self, path): self._global_path = Path(path)

    def set_config_dir_path(self, path): self._config_dir_path = Path(path)

    def set_exp_path(self, path): self._exp_path = Path(path)

    def set_config_file(self, filename): self._config_file_name = Path(filename)

    def get_rules_path (self):
        return self.get_global_path() / self._DEFAULT_RULES_FILE

    # 2019-11-23 RB TODO: Separate librec.jar and auto.jar files. Then restore the two jar classpaths here.
    def get_classpath (self):
        # return (self.get_global_path() / self.get_jar_path() / self._DEFAULT_LA_JAR).absolute().as_posix() + ";" + \
        #        (self.get_global_path() / self.get_jar_path() / self._DEFAULT_LR_JAR).absolute().as_posix()
        return (self.get_global_path() / self.get_jar_path() / self._DEFAULT_LA_JAR).absolute().as_posix()

    def get_subexp_name (self, count):
        return self._EXP_DIR_PATTERN.format(count)

    def detect_sub_path(self, exp_no):
        return (self.get_exp_path() / self.get_subexp_name(exp_no)).exists()

    def detect_sub_paths (self, count=0):
        sub_count = 0
        while self.detect_sub_path(sub_count):
            self._sub_path_dict[sub_count] = SubPaths(self.get_exp_path(), self.get_subexp_name(sub_count), create=False)
            sub_count += 1
        if sub_count != count:
            print(f'librec-auto: Expecting {count} existing experiment directories in {self.get_exp_path()}. Found {sub_count}.')

    def create_sub_paths (self, tuple_count):
        if tuple_count == 0:
            sub_exp_count = 1
        else:
            sub_exp_count = tuple_count

        for i in range(sub_exp_count):
            self._sub_path_dict[i] = SubPaths(self.get_exp_path(), self.get_subexp_name(i), create=True)

    def ensure_sub_paths(self, exp_count):
        if self.detect_sub_path(0):
            self.detect_sub_paths(count=exp_count)
        else:
            self.create_sub_paths(exp_count)

    def get_sub_count (self):
        if len(self._sub_path_dict) > 0:
            return max(self._sub_path_dict.keys())+1
        else:
            return 0

    def get_sub_paths(self, count):
        if count in self._sub_path_dict:
            return self._sub_path_dict[count]
        else:
            return None

    def get_sub_paths_iterator(self):
        return self._sub_path_dict.values()

    @staticmethod
    def dir_hash(maybe_path):
        """
        Starting from a directory, gets all subdirectories, extracts the individual files and creates a hash value
        using the file name, size, and last modification date.

        Probably just modification date is enough.
        :param maybe_path:
        :return:
        """
        hasher = hashlib.sha1()
        path = force_path(maybe_path)
        full_listing = path.glob('**/*')
        files = [fl for fl in full_listing if fl.is_file()]
        for fl in files:
            fl_stat = fl.stat()
            fl_size = fl_stat.st_size
            fl_date = fl_stat.st_mtime
            fl_info = "{}-{}-{}".format(fl.name, fl_size, fl_date)
            fl_bytes = fl_info.encode('utf-8')
            hasher.update(fl_bytes)
        return hasher.hexdigest()



from librec_auto.librec_auto.cmd import Cmd
from librec_auto.librec_auto.util import Files
from librec_auto.librec_auto.util.utils import create_param_spec
from librec_auto.librec_auto import ConfigCmd
import sys
import subprocess


class RerankCmd(Cmd):

    _files = None
    _config = None

    def __str__(self):
        return f'RerankCmd()'

    def setup(self, args):
        pass

    def dry_run(self, config):
        script_data = config.collect_scripts('rerank')[0]
        files = config.get_files()
        if files.get_sub_count() > 0:
            for i in range(0, files.get_sub_count()):
                sub_paths = files.get_sub_paths(i)

                print(f'librec-auto (DR): Running re-ranking command {self} for {sub_paths.subexp_name}')
                print(f'\tRe-rank script: {script_data[0]}')
                print(f'\tParameters: {script_data[1]}')
        else:
            print(f'librec-auto (DR): Running re-ranking command {self} but no results to re-rank.')
            print(f'    Re-rank script: {script}')

    # TODO RB 2019-12-12 You can only have one re-ranking script. Others will be silently ignored. Should probably
    # have a warning for this.
    def execute(self, config: ConfigCmd):
        self.status = Cmd.STATUS_INPROC
        self._files = config.get_files()
        self._config = config
        script_data = config.collect_scripts('rerank')[0]
        script = script_data[0]
        param_spec = create_param_spec(script_data[1])

        target = self._files.get_exp_path()

        if self._files.get_sub_count() > 0:
            for i in range(0, self._files.get_sub_count()):
                sub_paths = self._files.get_sub_paths(i)
                result_path = sub_paths.get_path('result')

                if not result_path.exists():
                    print('librec-auto: No results. Cannot rerank ', target)
                else:
                    print(f'librec-auto: Running re-ranking script {script} for {sub_paths.subexp_name}')
                    self.run_script(script, sub_paths, param_spec)

    def run_script(self, script, sub_paths, param_spec):
        # TODO: RB If rerank only, then leave original folder alone and delete results files
        # TODO: RB If config file is in a non-standard place, I think this will fail. Maybe the sub-scripts
        # should get passed the same information as the original script when invoked?
        sub_paths.results2original()
        proc_spec = [sys.executable, script.as_posix(), self._config.get_files().get_config_path().name,
                     self._config.get_target(),
                     sub_paths.get_path('original').absolute().as_posix(),
                     sub_paths.get_path('result').absolute().as_posix()] + param_spec
        print("    Parameters: " + str(proc_spec))
        subprocess.call(proc_spec)



from .sub_paths import SubPaths
from .files import Files
from .utils import safe_xml_path, extract_from_path, force_list, force_path, frange, \
    confirm, read_xml_from_path_string, xml_load_from_path, xml_load_from_text
from .log_file import LogFile
from .status import Status


from collections import OrderedDict
from librec_auto.librec_auto.util import xmltodict, Files, utils, xml_load_from_path
import logging
import itertools
from pathlib import Path
from librec_auto.librec_auto.config_lib import ConfigLibCollection, ConfigLib


class ConfigCmd:
    """
    Loads the configuration file using the element rules. Elements labels as "unparsed" in the rules are added to the
    _unparsed collection later processing.

    """

    _xml_input = None
    _rules_dict = None
    _prop_dict = None
    _var_data = None
    _unparsed = None
    _var_params = None
    _var_tuples = None
    _target = None
    _libraries = None

    IGNOREABLE_ATTRIBUTES = ['@ref', '@src', '@name']

    def __init__(self, config_file, target):

        self._files = Files()
        self._target = target

        # files.set_global_path(Path(__file__).parent.absolute())
        self._files.set_exp_path(target)
        self._files.set_config_file(config_file)

        self._xml_input = self.read_xml(self._files.get_config_path())

        self._rules_dict = self.read_rules()

        self._prop_dict = {}

        self._var_data = {}

        self._unparsed = {}

        self._var_params = []

        self._var_tuples = []

        self._libraries = ConfigLibCollection()

    def get_target(self):
        return self._target

    def get_prop_dict(self):
        return self._prop_dict

    def get_var_data(self):
        return self._var_data

    def get_value_tuple(self, subexp_no):
        return self._value_tuples[subexp_no]

    def get_sub_exp_count(self):
        exp_count = len(self._value_tuples)
        if exp_count == 0:
            return 1
        else:
            return exp_count

    # 2019-11-25 RB Configuration elements that aren't passed to LibRec are left in original XML format and handled
    # later in a command-specific way. A better way might be to have "handler" mechanism so that each command can
    # be associated its own configuration element and have code that is called when that element is encountered in the
    # parse. Something to think about.
    def get_unparsed(self, type):
        if type in self._unparsed:
            return self._unparsed[type]
        else:
            None

    def get_rules_dict(self):
        return self._rules_dict

    def get_files(self):
        return self._files

    def read_rules(self):
        rules_path = self._files.get_rules_path()
        if (rules_path.exists()):
            rules_input = xml_load_from_path(rules_path)
            return rules_input
        else:
            return None

    def read_xml(self, path_str):
        path = self._files.get_config_path()
        if (path.exists()):
            xml_input = xml_load_from_path(path)
            return xml_input
        else:
            return None

    def ensure_sub_experiments(self):
        exp_count = len(self._value_tuples)
        if exp_count == 0:
            exp_count = 1
        self.get_files().ensure_sub_paths(exp_count)

    def process_config(self):
        if type(self._xml_input) is OrderedDict:
            if type(self._rules_dict) is OrderedDict:
                self.setup_libraries()
                self.process_aux(self._xml_input['librec-auto'],
                                self._rules_dict['librec-auto-element-rules'])
                self.compute_value_tuples()
                self.ensure_sub_experiments()
            else:
                logging.error(f"Error processing element rules. Filename: {self._files.get_rules_path().as_posix()}")
        else:
            logging.error(f"Error processing configuration file. Filename: {self._files.get_config_path().as_posix()}")

    # Walk the configuration XML in tandem with the element rules to determine librec property translations for each
    # part of the configuration.
    def process_aux(self, arg, rules):
        for key in arg:
            if key in self.IGNOREABLE_ATTRIBUTES: # Skip bookkeeping attributes
                continue
            config_elem = self.augment_from_library(key, arg[key])
            if key in rules:                            # If the entry corresponds to a rule
                if "@action" in rules[key] and rules[key]['@action'] == 'no-parse': # If labeled "no parse" in rules
                    self._unparsed[key] = config_elem  # Add to unparsed collection
                elif type(config_elem) is OrderedDict:         # If the config file has subelements
                    if type(rules[key]) is OrderedDict:     # If the rules also have subelements
                        self.process_aux(config_elem, rules[key]) # recursive call
                    elif type(rules[key]) is list:          # If the rules have a list
                        self._prop_dict = self.process_attr(config_elem, rules[key])  # We have an attribute
                    elif 'value' in config_elem:               # If the config file has a 'value' key
                        self._var_data[rules[key]] = config_elem['value'] # then we have variable data for multiple exps.
                elif key in rules:                          # Config file doesn't have subelements
                    if type(config_elem) is list:              # Some properties have comma-separated values
                        self._prop_dict[rules[key]] = ','.join(config_elem)
                    elif type(rules[key]) == list:          # There are multiple LibRec keys in which map to this
                                                            # LibRecAuto key. (e.g. 'l1-reg')
                        for libRecKey in rules[key]:
                            if type(libRecKey) is str:            # Otherwise, it is a compound rule that doesn't match arg
                                self._prop_dict[libRecKey] = config_elem
                    else:
                        self._prop_dict[rules[key]] = config_elem  # Set property translation and value
            # If the key isn't in the rules, ignore it but warn because it is probably an error.
            else:
                logging.warning("Key {} is not in element rules.", key)

        return

    def augment_from_library(self, key, elem):
        if '@ref' not in elem:
            return elem
        else:
            ref = elem['@ref']
            lib_elem = self._libraries.get_element(ref).copy() # Don't want to overwrite library. Overly cautious?
            # Overwrite library entries with local
            if type(elem) is OrderedDict:
                for key,val in elem.items():
                    lib_elem[key] = val
            return lib_elem

    def get_string_rule(self, attr_rule):
        for item in attr_rule:
            if type(item) is str:
                    return item
        return None

    # Assumes attribute name is first in ordered dictionary.
    def collect_attributes(self, attr_rule):
        return [(list(item.keys())[0], item['#text'])
                for item in attr_rule if type(item) is OrderedDict]

    def process_attr(self, elem, attr_rule):
        # Scan rule for string
        # Associate with elem #text
        string_rule = self.get_string_rule(attr_rule)
        if 'value' in elem:                             # Variable rules
            self._var_data[string_rule] = elem['value']
        else:
            self._prop_dict[string_rule] = elem['#text']
        # Scan rule for all attributes
        # Assign
        attrib_pairs = self.collect_attributes(attr_rule)
        for attr_pair in attrib_pairs:
            self._prop_dict[attr_pair[1]] = elem[attr_pair[0]]
        return self._prop_dict

    def compute_value_tuples(self):
        self.var_params = self._var_data.keys()
        original_var_values = list(self._var_data.values())
        if (len(original_var_values) == 1):
            original_var_values = original_var_values[0]
        var_values = []
        for element in original_var_values:
            if type(element) is list:
                # print(element)
                var_values.append(element)
            else:
                var_values.append([element])
        if len(self.var_params) == 1:
            self._value_tuples = var_values
        else:
            self._value_tuples = list(itertools.product(*var_values))

    # TODO RB 2019-12-12 Should include some error-checking and better messages for badly-formed XML
    def collect_scripts(self, script_type):
        post_xml = self.get_unparsed(script_type)
        script_xml = post_xml['script']
        scripts = []
        for entry in utils.force_list(script_xml):
            script_path = utils.get_script_path(entry, script_type)
            if 'param' in entry:
                param_dict = {}
                for elem_dict in utils.force_list(entry['param']):
                    param_dict[elem_dict['@name']] = elem_dict['#text']
                scripts.append((script_path, param_dict))
            else:
                scripts.append((script_path, None))

        return scripts

    def setup_libraries(self):
        if utils.safe_xml_path(self._xml_input, ['librec-auto', 'library']):
            lib_elems = utils.force_list(utils.extract_from_path(self._xml_input, ['librec-auto', 'library']))
            for lib in lib_elems:
                lib_path = self.extract_library_path(lib)
                lib = ConfigLib(lib_path)
                self._libraries.append(lib)

    def extract_library_path(self, lib_elem):
        file_name = lib_elem['#text'] if type(lib_elem) is OrderedDict else lib_elem
        file_path = Path(file_name)
        path_prefix = None
        if type(lib_elem) is OrderedDict:
            if '@src' in lib_elem:
                if 'system' == lib_elem['@src']:
                    path_prefix = self._files.get_rules_path().parent
                else:
                    print(f'librec-auto: WARNING Path source {lib_elem["@src"]} is unknown. Possible values are: system')
        else: # If library path is just as string without directory information, assume conf directory
            if file_path.parent == Path('.'):
                path_prefix = self._files.get_config_path().parent
        if path_prefix is None:
            return file_path
        else:
            return path_prefix / file_path

def read_config_file(config_file, target):
    config = ConfigCmd(config_file, target)
    config.process_config()
    return config



import sys
import slack as sl
from slacker import Slacker

import argparse
from pathlib2 import Path

import matplotlib
matplotlib.use('Agg') # For non-windowed plotting
import matplotlib.pyplot as plt

def slack_send_message(message):
    #slack = Slacker('xoxb-625375728051-787189332448-Ufqx4usZgQUW9c3q3J0JsH7o')
    slack = Slacker(slack_cmd)
    slack.chat.post_message('#la_bot', message)

#Zijun: This is function for auto send file and comments to slack
def slack_send_file(rep, fil1, tit1):
    #client = sl.WebClient(token='xoxb-625375728051-787189332448-Ufqx4usZgQUW9c3q3J0JsH7o')
    client = sl.WebClient(token=slack_cmd)
    with open(rep, 'rb') as att:
        r = client.api_call("files.upload", files={'file': att,},
                            data={'channels': '#la_bot', 'filename': fil1, 'title': tit1, 'initial_comment': 'This is {}'.format(fil1),}
                            )
    assert r.status_code == 200

'''
Zijun:
    About how to run this file. 
    There are two action for the slack to recognize. The first one is the message. The second one is the file
    
    message:
        Description: If you want to direct post a sentence in to the slack channel. 
        The command should like this:
        python slack-post.py [action] [API-Token] ["Expression"]
        eg: python slack-post.py message xoxb-625375728051-787189332448-Ufqx4usZgQUW9c3q3J0JsH7o "Hello World"
        
    file: 
        Description: If you want to post file into slack.
        The command should like this:
        python slack-post.py [action] [API-Token] /directory
        eg: python slack-post.py file xoxb-625375728051-787189332448-Ufqx4usZgQUW9c3q3J0JsH7o /directory/---.py
        
'''


if __name__ == '__main__':
    slack_cmd = str(sys.argv[2])

    if len(sys.argv) < 2:
        print ("Argument input fail")

    parser = argparse.ArgumentParser()

    parser.add_argument("action", choices=["message", "file"])
    parser.add_argument("API-Token", type=str)
    parser.add_argument("content")

    args = parser.parse_args()
    dictargs = vars(args)

    if dictargs["action"] == "message":
        post_message = sys.argv[3]
        slack_send_message(post_message)
    if dictargs["action"] == "file":
        file_directory = sys.argv[3]
        slack_send_file(file_directory, "The file is here", "Plot")


